gem "activesupport"
require 'active_support/all'

module TestWise

  class ManualTestSuite 
    attr_accessor :name
    attr_accessor :test_cases

    def to_s
      the_str = ""
      the_str += "\nTest Suite: #{name}\n"
      @test_cases.each do |tc|
        the_str += tc.to_s
        the_str += "\n"
      end
      return the_str
    end

  end

  class ManualTestCase
    attr_accessor :pre_condition
    attr_accessor :description
    attr_accessor :test_steps

    attr_accessor :indent

    def initialize()
      @test_steps = []
      @indent = "  "      
    end

    def add_test_step(step)      
      
      if step =~ /^\s*expect\(\s*page_text\s*\)\.to\s+include\("(.*)"\)\s*$/
        @test_steps << "Expect \"#{$1}\" shown on the page"
      elsif step =~ /^\s*expect\(\s*driver\.page_source\s*\)\.to\s+include\("(.*)"\)\s*$/
        @test_steps << "Expect \"#{$1}\" in the page source"
      elsif step =~ /^\s*expect\(\s*page_source\s*\)\.to\s+include\("(.*)"\)\s*$/
        @test_steps << "Expect \"#{$1}\" in the page source"
      else
        @test_steps << step
      end
      
    end

    def to_s
      the_str = "\n" + indent + "TestCase #{@description}"
      the_str << "\n"
      @test_steps.each do |step|
        the_str << (@indent * 2 + step.to_s + "\n")
      end
      return the_str
    end

  end

end


module TestWise
  class SpecBehaviour

    attr_accessor :name
    attr_accessor :start_line, :end_line
    attr_accessor :spec_context

    attr_accessor :successful
    attr_accessor :duration
    attr_accessor :error
    attr_accessor :error_trace
    attr_accessor :lines_around_error

    def initialize(name, spec_file = nil)
      @name = name
      @spec_file = spec_file
      @lines_around_error = {}
    end


    def failed(failure)
      begin
        @error = failure.exception.message
        @error_trace = failure.exception.backtrace
        @lines_around_error[4] = "SSS"
        backtace_line_1 = @error_trace[0]
        if backtace_line_1 =~ /.*:(\d+):?\s*$/
          puts "line_No: #{$1}"
        end
      rescue => e
        puts "Failed to parse failed message"
      end

    end

    def get_test_case(name)
      my_name = name
      if @name =~ /^\s*'(.*)'\s*$/
        my_name = $1
      elsif @name =~ /^\s*"(.*)"\s*$/
        my_name = $1
      end

      if name == my_name
        return self
      else
        return nil
      end
    end

    def unquoted_name

      if @name =~ /^\s*'(.*)'\s*$/
        return $1
      elsif @name =~ /^\s*"(.*)"\s*$/
        return $1
      end
      @name
    end

    def display_name
      # the reason to +1, as the behaviour name contains double or single quotes,
      # where as in running 1 of xxx, the quotes will be stripped off.
      begin
        max_display_length = $prefs["max.testcase.name.display.length"].to_i rescue 36
        max_display_length = 36 if max_display_length < 10
      rescue => e
        max_display_length = 36
      end
      @name.length > max_display_length ? @name.strip[0, max_display_length + 1] + "..." : @name
    end

    def source
      @code = spec_file.lines[(start_line-1)..(end_line-1)].join("\n")
    end

    def spec_file
      @spec_context.spec_file rescue nil
    end

    def lines
      source ? source.split("\n") : []
    end

    def file_path
      @spec_context.file_path
    end

    def is_web_test?
      spec_file.is_web_test?
    end

  end
end


module TestWise

  class SpecContext

    attr_accessor :name
    attr_accessor :start_line
    attr_accessor :end_line
    attr_accessor :spec_behaviours

    attr_reader :spec_file

    attr_accessor :before_each_line_start, :before_all_line_start, :after_each_line_start, :after_all_line_start
    attr_accessor :first_test_case_line_start

    # 
    # before_each => [3,5]
    # "testset" => [2, 5]
    attr_accessor :blocks

    def initialize(name, spec_file)
      @name = name
      @spec_file = spec_file
      @spec_behaviours = []
      @blocks = {}
    end

    def display_name
      @name
    end

    def successful?
      @spec_behaviours.all? { |x| x.successful }
    end

    def block_start_line
      (@start_line+1..@end_line).each do |idx|
        line_text = @spec_file.lines[idx]
        if line_text =~ /^\s*include/ || line_text =~ /^\s*using/ || line_text =~ /^\s*extend/ || line_text =~ /^\s*pages\s/
          next
        end
        return idx
      end
    end

    def last_block_end_line
      block_end_lines = @blocks.values.collect { |x|
        begin
          x[1]
        rescue => e
        end
      }
      block_end_lines = block_end_lines.compact.flatten.sort
      begin
        return block_end_lines[-1]
      rescue => e
        return -1
      end
    end

    def get_test_case(name)
      @spec_behaviours.each { |x|
        testcase = x.get_test_case(name)
        return testcase if testcase
      }
      return nil
    end


    # eg. replace_block(:before_each, new_lines)
    def replace_block(name, lines, offset = 0)
      if @blocks[name] && lines.class == Array then
        # start line -1, inclusinve
        @spec_file.lines[@blocks[name][0]-1+offset..@blocks[name][1]-1+offset] = lines
        @spec_file.reparse
      end
    end

    def source
      lines.join()
    end

    def lines
      @spec_file.lines[(start_line-1)..(end_line-1)]
    end

    def file_path
      @spec_file.file_path
    end

    def normalize
      empty_lines_after_count = 0;
      @end_line.downto(@start_line) do |line_no|
        line = @spec_file.lines[line_no-1]
        if line.nil? || line.strip.empty?
          empty_lines_after_count += 1
        else
          break
        end
      end
      @end_line -= empty_lines_after_count
    end

    def first_test_case_start_line
      first_line = -1
      spec_behaviours.each do |test_case|
        first_line = test_case.start_line if test_case.start_line < first_line || first_line < 0
      end
      return first_line
    end

  end
end



module TestWise
  module RSpecManualExporter

    def generate_step_for_page_operation(page_operation, indent = 2)
      the_step = ""
      if page_operation =~ /^\s*(\w+)\s*\("(.*)"\)\s*$/
        function = $1
        param = $2
        the_step = "  #{function.humanize} \"#{param}\""
      elsif page_operation =~ /^\s*(\w+)\((.*)\)$/
        page_function_name = $1
        page_function_parameter = $2
        the_step = "  #{page_function_name.humanize} #{page_function_parameter}"
      else
        the_step = "  #{page_operation.humanize}"
      end
      return the_step
    end


    def export_manual
      suites = []
      @spec_contexts.each do |spec_context|
        new_suite = ManualTestSuite.new
        new_suite.name = spec_context.name

        spec_context.spec_behaviours.each do |behaviour|
          new_test_case = ManualTestCase.new
          new_test_case.description = behaviour.name

          behaviour.lines.each_with_index do |line, idx|
            next if idx == 0 # ignore first line

            if line =~ /([A-Z]\w+Page)\.new\(?@browser\)?\s*$/ || line =~ /([A-Z]\w+Page)\.new\(\s*driver\s*\)?\s*$/ || line =~ /([A-Z]\w+Page)\.new\(\s*browser\s*\)?\s*$/ 
              new_test_case.add_test_step("On #{$1.underscore.humanize}")
              next
            end
            
            if line =~ /^\s*(\w+)_page\.(.*)$/
              page_operation = $2.strip
              the_step = generate_step_for_page_operation(page_operation)
              new_test_case.add_test_step(the_step)
              next
            end

            if line =~ /^\s*sleep\s+(\d+)$/
              seconds = $1.to_i
              page_operation = "wait #{$1} second"
              page_operation += "s" if seconds > 1
              the_step = generate_step_for_page_operation(page_operation)
              new_test_case.add_test_step(the_step)
              next
            end

            if line =~ /^\s*try_until\s*\([\d|,|\s]+\)\s*\{(.*)\}\s*$/ || 
               line =~ /^\s*try_for\s*\([\d|,|\.\s]+\)\s*\{(.*)\}\s*$/
              try_operation = $1.strip
              if try_operation =~ /^(\w+)_page\.(.*)\s*$/ # page operation in try
                page_operation = $2.strip
                the_step = generate_step_for_page_operation(page_operation)
                new_test_case.add_test_step(the_step)
              else
                new_test_case.add_test_step(try_operation)
              end
              next
            end

            
            if line =~ /^\s*expect\(\s*page_text\s*\)\.to\s+include\("(.*)"\)\s*$/
              new_test_case.add_test_step("Expect \"#{$1}\" shown on the page")      
              next        
            end

            if line =~ /^\s*expect\(\s*driver\.page_source\s*\)\.to\s+include\("(.*)"\)\s*$/
              new_test_case.add_test_step("Expect \"#{$1}\" in the page source")      
              next        
            end

            if line =~ /^\s*expect\(\s*page_source\s*\)\.to\s+include\("(.*)"\)\s*$/
              new_test_case.add_test_step("Expect \"#{$1}\" in the page source")      
              next        
            end
            
            ## RWebSpec
            #
              
            if line =~ /^\s*handle_popup\(.*\)\s*\{(.*)\}\s*$/
              operation_invokes_popup = $1
              operation_invokes_popup.strip!
              line = operation_invokes_popup
            end

            if line =~ /^\s*expect_page (\w+)Page\s*$/
              new_test_case.add_test_step("Expect to be on #{$1.underscore.humanize} page")
              next
            end

            if line =~ /expect_page (\w+)Page\s*$/
              new_test_case.add_test_step("On #{$1.underscore.humanize} page")
              next
            end
            
            if line =~ /^\s*click_link\s*\("(.*)"\)\s*$/
              new_test_case.add_test_step("Click link \"#{$1}\"")
              next
            end

            if line =~ /^\s*click_button\s*\("(.*)"\)\s*$/
              new_test_case.add_test_step("Click button \"#{$1}\"")
              next
            end

            if line =~ /^\s*enter_text\s*\("(.*)"\s*,\s*\"(.*)\"\)\s*$/
              new_test_case.add_test_step("Enter \"#{$2}\" in text box '#{$1}'")
              next
            end

            if line =~ /^\s*click_radio_option\s*\("(.*)"\s*,\s*\"(.*)\"\)\s*$/
              new_test_case.add_test_step("Check radio option \"#{$2}\" for '#{$1}'")
              next
            end

            if line =~ /^\s*select_option\s*\("(.*)"\s*,\s*\"(.*)\"\)\s*$/
              new_test_case.add_test_step("Select \"#{$2}\" in drop down '#{$1}'")
              next
            end

            # Assertion
            if line =~ /^\s*page_text\.should\s*contain\("(.*)"\)\s*$/ || line =~ /^\s*assert_text_present\("(.*)"\)\s*$/
              new_test_case.add_test_step("Check the page constains text: \"#{$1}\"")
              next
            end

            if line =~ /^\s*page_text\.should_not\s*contain\("(.*)"\)\s*$/ || line =~ /^\s*assert_text_not_present\("(.*)"\)\s*$/
              new_test_case.add_test_step("Check the page does NOT constain text: \"#{$1}\"")
              next
            end

            if line =~ /^\s*page_source\.should\s*contain\("(.*)"\)\s*$/
              new_test_case.add_test_step("Check the page source constains html: \"#{$1}\"")
              next
            end

            if line =~ /^\s*page_source\.should_not\s*contain\("(.*)"\)\s*$/
              new_test_case.add_test_step("Check the page source does NOT constain html: \"#{$1}\"")
              next
            end

            if line =~ /^\s*assert_link_present_with_text\("(.*)"\)\s*$/
              new_test_case.add_test_step("Check the link: \"#{$1}\" present on page")
              next
            end

            if line =~ /^\s*assert_link_not_present_with_text\("(.*)"\)\s*$/
              new_test_case.add_test_step("Check the link: \"#{$1}\" not present on page")
              next
            end

           if line =~ /^\s*assert_hidden\((:\w+)\s*,\s*"(.*)"\)\s*$/
              new_test_case.add_test_step("Check the HTML fragment with ID: '#{$2}' not visible")
              next
            end

            next if line.strip.empty?
            next if line =~ /^end\s*$/ || line =~ /^\s\send\s*$/ || line =~ /^\tend\s*$/
            next if line.strip =~ /^\s*debug\s+/
            next if line.strip =~ /^\s*debug\s*\(/
            next if line.strip =~ /^#/

            new_test_case.add_test_step(line.strip)
          end

          suites << new_test_case
        end
      end

      return suites
    end # end export_manual

  end
end


module TestWise
  class SpecFile

    include RSpecManualExporter

    attr_accessor :file_path
    attr_accessor :spec_contexts
    attr_reader :lines
    attr_reader :file_name

    attr_reader :blocks
    
    def initialize(file_path, content_lines = nil)
      @spec_contexts = []
      @lines = []
      if content_lines then
        @lines = content_lines
      elsif file_path && File.exist?(file_path) then
        @file_path= file_path
        @file_name = File.basename(@file_path)
        begin
          File.open(@file_path) do |file|
            file.each_line { |line| @lines << line }
          end
        rescue => e
          puts ("Failed to open file: #{file_path}, #{e}")
        end
      end
      parse(@lines)
    end


    def get_test_case(name)
      @spec_contexts.each { |x|
        testcase = x.get_test_case(name)
        return testcase if testcase
      }
      return nil
    end


    def successful?
      @spec_contexts.all? { |x| x.successful? }
    end

    def test_cases
      the_test_cases = []
      @spec_contexts.each { |x|
        the_test_cases << x.spec_behaviours
      }
      the_test_cases.flatten!
    end

    def test_case_count
      the_count = 0
      @spec_contexts.each { |x|
        the_count += x.spec_behaviours.size
      }
      return the_count
    end

    def reparse
      @spec_contexts = []
      parse(@lines)
    end


    def parse(lines)
      cache_behaviours = []
      cache_contexts = []

      new_context = nil
      @current_block = nil
      lines.each_with_index { |line, idx|

        if new_context && line =~ /^\s*before\(:each\)\s+do\s*$/
          new_context.before_each_line_start = idx + 1
          new_context.blocks[:before_each] = [idx + 1]
          @current_block = new_context.blocks[:before_each]
        elsif new_context && line =~ /^\s*before\(:all\)\s+do\s*$/
          new_context.before_all_line_start= idx + 1
          new_context.blocks[:before_all] = [idx + 1]
          @current_block = new_context.blocks[:before_all]
        elsif new_context && line =~ /^\s*after\(:each\)\s+do\s*$/
          new_context.after_each_line_start= idx + 1
          new_context.blocks[:after_each] = [idx + 1]
          @current_block = new_context.blocks[:after_each]

        elsif new_context && line =~ /^\s*after\(:all\)\s+do\s*$/
          new_context.after_all_line_start= idx + 1
          new_context.blocks[:after_all] = [idx + 1]
          @current_block = new_context.blocks[:after_all]

        elsif  line =~ /^\s*(context|describe|specification|test_suite|suite)(.*)do/

          context_name = $2.strip
          new_context = SpecContext.new(context_name, self)
          new_context.start_line = idx + 1
          cache_contexts << new_context
        elsif new_context && line =~ /^\s*(specify|it|story|scenario|test_case|use_case|test)(.*)do/

          testcase_name = $2.strip
          new_behaviour = SpecBehaviour.new(testcase_name, self)
          new_behaviour.start_line = idx + 1
          if new_behaviour
            new_context.first_test_case_line_start = new_behaviour.start_line unless new_context.first_test_case_line_start
          end
          cache_behaviours << new_behaviour
        elsif line =~ /^\s*end\s*$/

          if @current_block
            @current_block << idx + 1
            @current_block.flatten!
            @current_block = nil
          end
        end
      }

      cache_behaviours.reverse.each { |behaviour|
        cache_contexts.reverse.each { |context|
          if behaviour.start_line > context.start_line
            context.spec_behaviours << behaviour
            behaviour.spec_context = context
            break
          end
        }
      }

      cache_contexts.each { |ctx| ctx.spec_behaviours.reverse!; @spec_contexts << ctx }

      last_ctx = nil
      @spec_contexts.reverse.each_with_index { |ctx, idx|

        if idx == 0
          ctx.end_line = lines.size # will using -1 for range
          last_ctx = ctx
          next
        end

        ctx.end_line = last_ctx.start_line - 1
        last_ctx = ctx
      }

      last_beh = nil
      @spec_contexts.each_with_index { |ctx, idx|
        ctx.spec_behaviours.reverse.each_with_index { |beh, bdx|
          if bdx == 0 then
            beh.end_line = ctx.end_line - 1
            last_beh = beh
            next
          end

          beh.end_line = last_beh.start_line - 1
          last_beh = beh
        }
      }

      # remove empty lines after last 'end'
      normalize

      @blocks = []
      @spec_contexts.each { |x|
        @blocks << x.blocks if x.blocks && !x.blocks.empty?
      }

      @blocks.flatten!
    end

    def to_s
      str = "File :#{file_path}\n"
      @spec_contexts.each { |ctx|
        str << "  context: #{ctx.name} - (#{ctx.start_line} : #{ctx.end_line})\n"
        ctx.spec_behaviours.each { |beh|
          str << "    behaviour: #{beh.name} - (#{beh.start_line} : #{beh.end_line})\n"
        }
      }
      str
    end


    def behaviour_lookups
      lookups = {}
      @spec_contexts.each { |ctx|
        ctx.spec_behaviours.each { |beh|
          lookups[beh.name] = beh
        }
      }
      return lookups
    end

    def get_behaviour_by_line(line_number)
      @spec_contexts.each { |ctx|
        if (line_number >= ctx.start_line && line_number <= ctx.end_line) then
          ctx.spec_behaviours.each { |beh|
            return beh if line_number >= beh.start_line && line_number <= beh.end_line
          }
          # not in contained the behaviour,
          return ctx
        end
      }
      # not found, return self
      return self
    end

    def name
      @file_name
    end

    alias display_name name

    def start_line
      1
    end

    def end_line
      lines.size
    end

    def is_web_test?
      spec_content = @lines.join("")
      spec_content.include?("RWebUnit::") || spec_content.include?("RWebSpec::") || spec_content.include?("open_browser")
    end


    # clean up lines
    def normalize
      @spec_contexts.each { |ctx|
        ctx.normalize
      }
    end


    def get_spec_context_by_line(line)
      @spec_contexts.each { |ctx|
        return ctx if line < ctx.end_line && line > ctx.start_line
      }
      return nil
    end


  end

end

if ARGV.count == 1 
  test_file = ARGV[0]
  if File.exist?(test_file)
    spec_file =::TestWise::SpecFile.new(test_file)
    puts spec_file.export_manual
  end
else  
  exit(-1)
end

