# frozen_string_literal: true

# for Cucumber 4
require 'cucumber/formatter/backtrace_filter'
require 'cucumber/formatter/console'
require 'cucumber/formatter/console_counts'
require 'cucumber/formatter/console_issues'
require 'cucumber/formatter/io'
require 'cucumber/formatter/duration_extractor'
require 'cucumber/formatter/ast_lookup'
require 'cucumber/formatter/interceptor'

require "#{File.dirname(__FILE__)}/testwise_listener.rb"
require "#{File.dirname(__FILE__)}/testwise_runtime_support.rb"

class TestWiseCucumberFormatter
  
      include ::Cucumber::Formatter::Console
      include ::Cucumber::Formatter::Io
      
      include TestWiseRuntimeSupport
      
      attr_reader :config, :current_feature_uri
      private :config, :current_feature_uri

      def initialize(config)
        @config = config
#        @io = ensure_io(config.out_stream)
        @snippets_input = []
        @undefined_parameter_types = []
        @total_duration = 0
        @matches = {}
        @pending_step_matches = []
        @failed_results = []
        @passed_test_cases = []
        @current_feature_uri = ''
        @gherkin_documents = {}
        # @ast_lookup = AstLookup.new(config)
        # @counts = ConsoleCounts.new(config)
        # @issues = ConsoleIssues.new(config, @ast_lookup)
        config.on_event :test_run_started, &method(:on_test_run_started)
        config.on_event :step_activated, &method(:on_step_activated)
        config.on_event :test_case_started, &method(:on_test_case_started)
        config.on_event :test_step_finished, &method(:on_test_step_finished)
        config.on_event :test_case_finished, &method(:on_test_case_finished)
        config.on_event :test_run_finished, &method(:on_test_run_finished)
        config.on_event :undefined_parameter_type, &method(:collect_undefined_parameter_type_names)
                
        ::TestWise::Runtime.load_webdriver_support
        ::TestWise::Runtime.load_watir_support
      end

      def on_test_run_started(event)
        hash = {}
        # hash[:group_name] = feature.name
        hash[:test_case_count] = event.test_cases.size
        json_data = hash.to_json
        connect_to_testwise("    START",  json_data + "\r\n")            
      end
      
      def on_step_activated(event)
        @matches[event.test_step.to_s] = event.step_match
      end

      def on_test_case_started(event)
        unless @profile_information_printed
          do_print_profile_information(config.profiles) unless config.skip_profile_information? || config.profiles.nil? || config.profiles.empty?
          @profile_information_printed = true
        end
        @current_feature_uri = event.test_case.location.file
        
        # TestWise Support
        hash = {}
        hash[:file_path] = @current_feature_uri   
        # /Users/zhimin/work/projects/testwise-cpp/samples/demo-page-objects/features/01_login.feature:12
        hash[:test_case] = event.test_case.name        
        json_data = hash.to_json        
        connect_to_testwise("TESTSTART",  json_data + "\r\n")  
        
        # In order to fill out <system-err/> and <system-out/>, we need to
        # intercept the $stderr and $stdout
        @interceptedout = ::Cucumber::Formatter::Interceptor::Pipe.wrap(:stdout)
        @interceptederr = ::Cucumber::Formatter::Interceptor::Pipe.wrap(:stderr) 
      end

      def on_test_step_finished(event)
        test_step = event.test_step
        result = event.result.with_filtered_backtrace(Cucumber::Formatter::BacktraceFilter)
        progress(result.to_sym) if !test_step.hook? || result.failed?

        return if test_step.hook?
        # collect_snippet_data(test_step, @ast_lookup) if result.undefined?
        @pending_step_matches << @matches[test_step.to_s] if result.pending?
        @failed_results << result if result.failed?        
        
        ::Cucumber::Formatter::Interceptor::Pipe.unwrap! :stdout
        ::Cucumber::Formatter::Interceptor::Pipe.unwrap! :stderr
      end
            
      def on_test_case_finished(event)
        test_case = event.test_case
        result = event.result.with_filtered_backtrace(Cucumber::Formatter::BacktraceFilter)
        @passed_test_cases << test_case if result.passed?
        @total_duration += ::Cucumber::Formatter::DurationExtractor.new(result).result_duration
        
        hash = {}
        hash[:file_path] = test_case.location.file
        hash[:test_case] = test_case.name
        json_data = hash.to_json

        if result.class.name =~ /Passed/
            connect_to_testwise(" TESTPASS",  json_data + "\r\n")
        else
            # puts result.exception.inspect
            # puts result.exception.backtrace.inspect
            error_message = result.exception.message
            error_backtrace = result.exception.backtrace
            
            # prevent too big
            if error_message.size > 2000
                error_message = error_message[0..1744] + "..." + error_message[-256..-1]
            end
            hash[:error_message] = error_message
            
            if error_backtrace.size > 2000
                error_backtrace = error_backtrace[0..1000] + "..." + error_backtrace[-1000..-1]
            end
            hash[:error_backtrace] = error_backtrace
            
            json_data = hash.to_json
            connect_to_testwise(" TESTFAIL",  json_data + "\r\n")
        end

              
      end

      def on_test_run_finished(_event)
        # @io.puts
        # @io.puts
        #print_summary
      end

      private

      def gherkin_document
        @ast_lookup.gherkin_document(current_feature_uri)
      end

      def print_summary
        print_elements(@pending_step_matches, :pending, 'steps')
        print_elements(@failed_results, :failed, 'steps')
        print_statistics(@total_duration, @config, @counts, @issues)
        print_snippets(config.to_hash)
        print_passing_wip(config, @passed_test_cases, @ast_lookup)
      end

      CHARS = {
        passed: '.',
        failed: 'F',
        undefined: 'U',
        pending: 'P',
        skipped: '-'
      }.freeze

      def progress(status)
        char = CHARS[status]
        # @io.print(format_string(char, status))
        # @io.flush
      end

      def table_header_cell?(status)
        status == :skipped_param
      end

      TestCaseData = Struct.new(:name, :location)
    end
